//
use color_eyre::Result;
use kameo::message::{Context, Message};
use tracing::warn;

use git_next_core::git;

use crate::{
    repo::{
        branch::{advance_next, find_next_commit_on_dev},
        do_send,
        messages::{AdvanceNext, AdvanceNextPayload, ValidateRepo},
        RepoActor,
    },
    server::actor::messages::RepoUpdate,
};

impl Message<AdvanceNext> for RepoActor {
    type Reply = Result<()>;

    async fn handle(
        &mut self,
        msg: AdvanceNext,
        ctx: Context<'_, Self, Self::Reply>,
    ) -> Self::Reply {
        let Some(repo_config) = &self.repo_details.repo_config else {
            return Ok(());
        };
        let Some(open_repository) = &self.open_repository else {
            return Ok(());
        };

        let AdvanceNextPayload {
            next,
            main,
            dev_commit_history,
        } = msg.peel();
        let (commit, force) = find_next_commit_on_dev(&next, &main, &dev_commit_history);
        if let Some(commit) = &commit {
            self.update_tui(RepoUpdate::AdvancingNext {
                commit: commit.clone(),
                force: force.clone(),
            })
            .await?;
        };
        match advance_next(
            commit,
            force,
            &self.repo_details,
            repo_config,
            &**open_repository,
            self.message_token,
        ) {
            Ok(message_token) => {
                self.update_tui(RepoUpdate::NextUpdated).await?;
                match open_repository.fetch() {
                    Ok(()) => {
                        self.update_tui_log(git::graph::log(&self.repo_details))
                            .await?;
                    }
                    Err(err) => self.alert_tui(format!("fetching: {err}")).await?,
                }
                // INFO: pause to allow any CI checks to be started
                tokio::time::sleep(self.sleep_duration).await;
                Ok(do_send(
                    &ctx.actor_ref(),
                    ValidateRepo::new(message_token),
                    self.log.as_ref(),
                )
                .await?)
            }
            Err(err) => {
                warn!("advance next: {err}");
                self.alert_tui(err.to_string()).await?;
                Ok(())
            }
        }
    }
}
