use crate::{
    alerts::{AlertsActor, History},
    server::{actor::messages::ServerUpdate, ServerActor},
};

//
use super::*;

use git_next_core::server::ListenUrl;
use kameo::actor::{pubsub::PubSub, ActorRef};
use kxio::{fs::FileSystem, net::Net};

pub fn has_all_valid_remote_defaults(
    open_repository: &mut MockOpenRepositoryLike,
    repo_details: &RepoDetails,
) {
    has_remote_defaults(
        open_repository,
        HashMap::from([
            (Direction::Push, repo_details.remote_url()),
            (Direction::Fetch, repo_details.remote_url()),
        ]),
    );
}

pub fn has_remote_defaults(
    open_repository: &mut MockOpenRepositoryLike,
    remotes: HashMap<Direction, Option<RemoteUrl>>,
) {
    for (direction, remote) in remotes {
        open_repository
            .expect_find_default_remote()
            .with(eq(direction))
            .return_once(|_| remote);
    }
}

pub fn a_webhook_auth() -> WebhookAuth {
    WebhookAuth::generate()
}

pub fn repo_branches() -> RepoBranches {
    RepoBranches::new(
        format!("main-{}", a_name()),
        format!("next-{}", a_name()),
        format!("dev-{}", a_name()),
    )
}

pub fn a_forge_alias() -> ForgeAlias {
    ForgeAlias::new(a_name())
}

pub fn a_repo_alias() -> RepoAlias {
    RepoAlias::new(a_name())
}

pub fn a_network() -> kxio::net::MockNet {
    kxio::net::mock()
}

pub fn a_listen_url() -> ListenUrl {
    ListenUrl::new(a_name())
}

pub fn a_name() -> String {
    use rand::Rng;
    use std::iter;

    fn generate(len: usize) -> String {
        const CHARSET: &[u8] = b"ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789";
        let mut rng = rand::thread_rng();
        let one_char = || CHARSET[rng.gen_range(0..CHARSET.len())] as char;
        iter::repeat_with(one_char).take(len).collect()
    }
    generate(7)
}

pub fn maybe_a_number() -> Option<u32> {
    use rand::Rng;
    let mut rng = rand::thread_rng();
    if Rng::gen_ratio(&mut rng, 1, 2) {
        Some(a_number())
    } else {
        None
    }
}

pub fn a_number() -> u32 {
    use rand::Rng;
    let mut rng = rand::thread_rng();
    rng.gen_range(0..100)
}

pub fn a_webhook_id() -> WebhookId {
    WebhookId::new(a_name())
}

pub fn a_branch_name(prefix: impl Into<String>) -> BranchName {
    BranchName::new(format!("{}-{}", prefix.into(), a_name()))
}

pub fn a_git_dir(fs: &kxio::fs::FileSystem) -> GitDir {
    let dir_name = a_name();
    let dir = fs.base().join(dir_name);
    GitDir::new(dir, StoragePathType::Internal)
}

pub fn a_forge_config() -> ForgeConfig {
    ForgeConfig::new(
        ForgeType::MockForge,
        a_name(),
        a_name(),
        a_name(),
        maybe_a_number(),
        BTreeMap::default(), // no repos
    )
}

pub fn a_server_repo_config() -> ServerRepoConfig {
    let main = a_branch_name("main").to_string();
    let next = a_branch_name("next").to_string();
    let dev = a_branch_name("dev").to_string();
    ServerRepoConfig::new(
        format!("{}/{}", a_name(), a_name()),
        main.clone(),
        None,
        Some(main),
        Some(next),
        Some(dev),
    )
}

pub fn a_repo_config() -> RepoConfig {
    RepoConfig::new(given::repo_branches(), RepoConfigSource::Repo)
}

pub fn a_named_commit(name: impl Into<String>) -> Commit {
    Commit::new(a_named_commit_sha(name), a_commit_message())
}

pub fn a_commit() -> Commit {
    Commit::new(a_commit_sha(), a_commit_message())
}

pub fn a_commit_with_message(message: impl Into<crate::git::commit::Message>) -> Commit {
    Commit::new(a_commit_sha(), message.into())
}

pub fn a_commit_message() -> crate::git::commit::Message {
    crate::git::commit::Message::new(a_name())
}

pub fn a_named_commit_sha(name: impl Into<String>) -> Sha {
    Sha::new(format!("{}-{}", name.into(), a_name()))
}

pub fn a_commit_sha() -> Sha {
    Sha::new(a_name())
}

pub fn a_filesystem() -> kxio::fs::TempFileSystem {
    #[allow(clippy::expect_used)]
    kxio::fs::temp().expect("temp fs")
}

pub fn repo_details(fs: &kxio::fs::FileSystem) -> RepoDetails {
    let generation = Generation::default();
    let repo_alias = a_repo_alias();
    let server_repo_config = a_server_repo_config();
    let forge_alias = a_forge_alias();
    let forge_config = a_forge_config();
    let gitdir = a_git_dir(fs);
    RepoDetails::new(
        generation,
        &repo_alias,
        &server_repo_config,
        &forge_alias,
        &forge_config,
        gitdir,
    )
}

pub fn an_open_repository(fs: &kxio::fs::FileSystem) -> (MockOpenRepositoryLike, RepoDetails) {
    let open_repository = MockOpenRepositoryLike::new();
    let gitdir = given::a_git_dir(fs);
    let hostname = given::a_hostname();
    let repo_details = given::repo_details(fs)
        .with_gitdir(gitdir)
        .with_hostname(hostname);
    (open_repository, repo_details)
}

pub fn a_message_token() -> MessageToken {
    MessageToken::default()
}

#[allow(clippy::unnecessary_box_returns)]
pub fn a_forge() -> Box<MockForgeLike> {
    Box::new(MockForgeLike::new())
}

pub fn a_repo_actor(
    repo_details: RepoDetails,
    repository_factory: Box<dyn RepositoryFactory>,
    forge: Box<dyn ForgeLike>,
    server_actor: ActorRef<ServerActor>,
    net: kxio::net::Net,
) -> (RepoActor, ActorLog) {
    let listen_url = given::a_listen_url();
    let generation = Generation::default();
    let log = ActorLog::default();
    let actors_log = log.clone();
    (
        RepoActor::new(
            false,
            repo_details,
            forge,
            listen_url,
            generation,
            net,
            repository_factory,
            std::time::Duration::from_nanos(0),
            None,
            server_actor,
        )
        .with_log(actors_log),
        log,
    )
}

pub fn a_server_actor(fs: FileSystem, net: Net) -> ActorRef<ServerActor> {
    let alerts_actor_ref = kameo::spawn(AlertsActor::new(
        None,
        History::new(Duration::from_secs(0)),
        net.clone(),
    ));
    let server_updates_actor_ref = kameo::spawn(PubSub::<ServerUpdate>::new());
    let repo_factor = Box::new(MockRepositoryFactory::new());
    let duration = Duration::from_secs(0);
    let server_actor = ServerActor::new(
        false, // ui
        fs,
        net,
        alerts_actor_ref,
        server_updates_actor_ref,
        repo_factor,
        duration,
    );
    kameo::spawn(server_actor)
}

pub fn a_hostname() -> Hostname {
    Hostname::new(given::a_name())
}

pub fn a_registered_webhook() -> RegisteredWebhook {
    RegisteredWebhook::new(given::a_webhook_id(), given::a_webhook_auth())
}

pub fn a_push() -> Push {
    Push::new(
        given::a_branch_name("push"),
        given::a_name(),
        given::a_name(),
    )
}
