use kxio::net::Net;

use crate::tell;

//
use super::*;

#[test_log::test(tokio::test)]
async fn should_clone() -> TestResult {
    //given
    let fs = given::a_filesystem();
    let (mut open_repository, /* mut */ repo_details) = given::an_open_repository(&fs);
    // #[allow(clippy::unwrap_used)]
    // let repo_config = repo_details.repo_config.take().unwrap();

    given::has_all_valid_remote_defaults(&mut open_repository, &repo_details);
    // handles_validate_repo_message(&mut open_repository, repo_config.branches());

    // factory clones an open repository
    let mut repository_factory = MockRepositoryFactory::new();
    let cloned = Arc::new(RwLock::new(vec![]));
    let cloned_ref = cloned.clone();
    repository_factory
        .expect_git_clone()
        .times(2)
        .return_once(move |_| {
            let _ = cloned_ref.write().map(|mut l| l.push(()));
            Ok(Box::new(open_repository))
        });
    let net: Net = given::a_network().into();

    //when
    let (addr, _log) = when::start_actor(
        repository_factory,
        repo_details,
        given::a_forge(),
        fs.as_real(),
        net.clone(),
    );
    tell!(addr, CloneRepo::new())?;
    //then
    tick(1).await;
    cloned
        .read()
        .map_err(|e| e.to_string())
        .map(|o| assert_eq!(o.len(), 1))?;
    net.assert_no_unused_plans();

    Ok(())
}

#[tokio::test]
async fn should_open() -> TestResult {
    //given
    let fs = given::a_filesystem();
    let (mut open_repository, repo_details) = given::an_open_repository(&fs);
    open_repository
        .expect_fetch()
        .times(1)
        .return_once(|| Ok(()));
    given::has_all_valid_remote_defaults(&mut open_repository, &repo_details);
    // factory opens a repository
    let mut repository_factory = MockRepositoryFactory::new();
    let opened = Arc::new(RwLock::new(vec![]));
    let opened_ref = opened.clone();
    repository_factory
        .expect_open()
        .times(1)
        .return_once(move |_| {
            let _ = opened_ref.write().map(|mut l| l.push(()));
            Ok(Box::new(open_repository))
        });
    fs.dir(&repo_details.gitdir).create()?;
    let net: Net = given::a_network().into();

    //when
    let (addr, _log) = when::start_actor(
        repository_factory,
        repo_details,
        given::a_forge(),
        fs.as_real(),
        net.clone(),
    );
    tell!(addr, CloneRepo::new())?;

    //then
    tick(1).await;
    opened
        .read()
        .map_err(|e| e.to_string())
        .map(|o| assert_eq!(o.len(), 1))?;
    net.assert_no_unused_plans();

    Ok(())
}

/// The server config can optionally include the names of the main, next and dev
/// branches. When it doesn't we should load the `.git-next.yaml` from from the
/// repo and get the branch names from there by sending a [LoadConfigFromRepo] message.
#[tokio::test]
async fn when_server_has_no_repo_config_should_send_load_from_repo() -> TestResult {
    //given
    let fs = given::a_filesystem();
    let (mut open_repository, mut repo_details) = given::an_open_repository(&fs);
    open_repository
        .expect_fetch()
        .times(1)
        .return_once(|| Ok(()));
    #[allow(clippy::unwrap_used)]
    let _repo_config = repo_details.repo_config.take().unwrap();

    given::has_all_valid_remote_defaults(&mut open_repository, &repo_details);

    let mut repository_factory = MockRepositoryFactory::new();
    expect::open_repository(&mut repository_factory, open_repository);
    fs.dir(&repo_details.gitdir).create()?;
    let net: Net = given::a_network().into();

    //when
    let (addr, log) = when::start_actor(
        repository_factory,
        repo_details,
        given::a_forge(),
        fs.as_real(),
        net.clone(),
    );
    tell!(addr, CloneRepo::new())?;

    //then
    log.require_message_containing("send: LoadConfigFromRepo")
        .await?;
    net.assert_no_unused_plans();

    Ok(())
}

/// The server config can optionally include the names of the main, next and dev
/// branches. When it does we should register the webhook by sending [RegisterWebhook] message.
#[test_log::test(tokio::test)]
async fn when_server_has_repo_config_should_send_register_webhook() -> TestResult {
    //given
    let fs = given::a_filesystem();
    let (mut open_repository, repo_details) = given::an_open_repository(&fs);
    open_repository
        .expect_fetch()
        .times(1)
        .return_once(|| Ok(()));
    #[allow(clippy::unwrap_used)]
    given::has_all_valid_remote_defaults(&mut open_repository, &repo_details);

    let mut repository_factory = MockRepositoryFactory::new();
    expect::open_repository(&mut repository_factory, open_repository);
    fs.dir(&repo_details.gitdir).create()?;
    let net: Net = given::a_network().into();

    //when
    let (addr, log) = when::start_actor(
        repository_factory,
        repo_details,
        given::a_forge(),
        fs.as_real(),
        net.clone(),
    );
    tell!(addr, CloneRepo::new())?;

    //then
    tick(1).await;
    debug!(?log, "");
    log.require_message_containing("send: RegisterWebhook")
        .await?;
    net.assert_no_unused_plans();

    Ok(())
}
