//
use kxio::net::Net;

use crate::{
    repo::messages::{AdvanceNext, AdvanceNextPayload, ValidateRepo},
    tell,
};

use super::*;

#[test_log::test(tokio::test)]
async fn repo_with_next_not_an_ancestor_of_dev_and_dev_on_main_should_be_reset_to_main(
) -> TestResult {
    //given
    let fs = given::a_filesystem();
    let (mut open_repository, repo_details) = given::an_open_repository(&fs);
    #[allow(clippy::unwrap_used)]
    let repo_config = repo_details.repo_config.clone().unwrap();
    expect::fetch_ok(&mut open_repository);
    let branches = repo_config.branches();
    // commit_log main
    let main_commit = expect::main_commit_log(&mut open_repository, branches.main());
    // next - based on main
    let next_branch_log = vec![given::a_commit(), main_commit.clone()];
    // dev - based on main, but not on next
    let dev_branch_log = vec![main_commit.clone()];
    // commit_log next - based on main, but not a parent of dev
    open_repository
        .expect_commit_log()
        .times(1)
        .with(eq(branches.next()), eq([main_commit.clone()]))
        .return_once(move |_, _| Ok(next_branch_log));
    // commit_log dev
    open_repository
        .expect_commit_log()
        .times(1)
        .with(eq(branches.dev()), eq([main_commit]))
        .return_once(|_, _| Ok(dev_branch_log));
    // expect to reset the branch
    expect::fetch_ok(&mut open_repository);
    expect::push_ok(&mut open_repository);

    //when
    let (addr, log) = when::start_actor_with_open_repository(
        Box::new(open_repository),
        repo_details,
        given::a_forge(),
    );
    tell!(addr, ValidateRepo::new(MessageToken::default()))?;

    //then
    log.require_message_containing(format!("Branch {} has been reset", branches.next()))
        .await?;
    Ok(())
}

#[test_log::test(tokio::test)]
async fn repo_with_next_not_an_ancestor_of_dev_and_dev_ahead_of_main_should_be_reset_to_dev(
) -> TestResult {
    //given
    let fs = given::a_filesystem();
    let (mut open_repository, repo_details) = given::an_open_repository(&fs);
    #[allow(clippy::unwrap_used)]
    let repo_config = repo_details.repo_config.clone().unwrap();
    expect::fetch_ok(&mut open_repository);
    let branches = repo_config.branches();
    // commit_log main
    let main_commit = expect::main_commit_log(&mut open_repository, branches.main());
    // next - based on main
    let next_commit = given::a_commit();
    let next_branch_log = vec![next_commit.clone(), main_commit.clone()];
    // dev - based on main, but not on next
    let dev_branch_log = vec![given::a_commit(), main_commit.clone()];
    // commit_log next - based on main, but not a parent of dev
    open_repository
        .expect_commit_log()
        .times(1)
        .with(eq(branches.next()), eq([main_commit.clone()]))
        .return_once(move |_, _| Ok(next_branch_log));
    // commit_log dev
    let dev_branch_log_clone = dev_branch_log.clone();
    open_repository
        .expect_commit_log()
        .times(1)
        .with(eq(branches.dev()), eq([main_commit.clone()]))
        .return_once(|_, _| Ok(dev_branch_log_clone));
    // expect to reset the branch
    expect::fetch_ok(&mut open_repository);
    expect::push_ok(&mut open_repository);

    //when
    let (addr, log) = when::start_actor_with_open_repository(
        Box::new(open_repository),
        repo_details,
        given::a_forge(),
    );
    tell!(
        addr,
        crate::repo::messages::ValidateRepo::new(MessageToken::default())
    )?;

    //then
    let expected = AdvanceNext::new(AdvanceNextPayload {
        next: next_commit,
        main: main_commit,
        dev_commit_history: dev_branch_log,
    });
    log.require_message_containing(format!("send: {expected:?}",))
        .await?;
    Ok(())
}

#[test_log::test(tokio::test)]
async fn repo_with_next_not_on_or_near_main_should_be_reset() -> TestResult {
    //given
    let fs = given::a_filesystem();
    let (mut open_repository, repo_details) = given::an_open_repository(&fs);
    #[allow(clippy::unwrap_used)]
    let repo_config = repo_details.repo_config.clone().unwrap();
    expect::fetch_ok(&mut open_repository);
    let branches = repo_config.branches();
    // commit_log main
    let main_commit = expect::main_commit_log(&mut open_repository, branches.main());
    // next - based on main, but too far in advance
    let next_branch_log = vec![given::a_commit(), given::a_commit(), main_commit.clone()];
    // dev - based on next
    let mut dev_branch_log = vec![given::a_commit()];
    dev_branch_log.extend(next_branch_log.clone());
    // commit_log next
    open_repository
        .expect_commit_log()
        .times(1)
        .with(eq(branches.next()), eq([main_commit.clone()]))
        .return_once(move |_, _| Ok(next_branch_log));
    // commit_log dev
    open_repository
        .expect_commit_log()
        .times(1)
        .with(eq(branches.dev()), eq([main_commit]))
        .return_once(|_, _| Ok(dev_branch_log));
    expect::fetch_ok(&mut open_repository);
    expect::push_ok(&mut open_repository);

    //when
    let (addr, log) = when::start_actor_with_open_repository(
        Box::new(open_repository),
        repo_details,
        given::a_forge(),
    );
    tell!(addr, ValidateRepo::new(MessageToken::default()))?;

    //then
    log.require_message_containing(format!("Branch {} has been reset", branches.next()))
        .await?;
    Ok(())
}

#[test_log::test(tokio::test)]
async fn repo_with_next_not_based_on_main_should_be_reset() -> TestResult {
    //given
    let fs = given::a_filesystem();
    let (mut open_repository, repo_details) = given::an_open_repository(&fs);
    #[allow(clippy::unwrap_used)]
    let repo_config = repo_details.repo_config.clone().unwrap();
    // given::has_all_valid_remote_defaults(&mut open_repository, &repo_details);
    expect::fetch_ok(&mut open_repository);
    let branches = repo_config.branches();
    // commit_log main
    let main_commit = expect::main_commit_log(&mut open_repository, branches.main());
    // next - not based on main
    let next_branch_log = vec![given::a_commit(), given::a_commit(), given::a_commit()];
    // dev - based on main
    let dev_branch_log = vec![given::a_commit(), main_commit.clone()];
    // commit_log next
    open_repository
        .expect_commit_log()
        .times(1)
        .with(eq(branches.next()), eq([main_commit.clone()]))
        .return_once(move |_, _| Ok(next_branch_log));
    // commit_log dev
    open_repository
        .expect_commit_log()
        .times(1)
        .with(eq(branches.dev()), eq([main_commit]))
        .return_once(|_, _| Ok(dev_branch_log));
    expect::fetch_ok(&mut open_repository);
    expect::push_ok(&mut open_repository);

    //when
    let (addr, log) = when::start_actor_with_open_repository(
        Box::new(open_repository),
        repo_details,
        given::a_forge(),
    );
    tell!(addr, ValidateRepo::new(MessageToken::default()))?;

    //then
    log.require_message_containing(format!("Branch {} has been reset", branches.next()))
        .await?;
    Ok(())
}

#[test_log::test(tokio::test)]
async fn repo_with_next_ahead_of_main_should_check_ci_status() -> TestResult {
    //given
    let fs = given::a_filesystem();
    let (mut open_repository, repo_details) = given::an_open_repository(&fs);
    #[allow(clippy::unwrap_used)]
    let repo_config = repo_details.repo_config.clone().unwrap();
    // Validate repo branches
    expect::fetch_ok(&mut open_repository);
    let branches = repo_config.branches();
    // commit_log main
    let main_commit = expect::main_commit_log(&mut open_repository, branches.main());
    // next - ahead of main
    let next_commit = given::a_named_commit("next");
    let next_branch_log = vec![next_commit.clone(), main_commit.clone()];
    // dev - on next
    let dev_branch_log = next_branch_log.clone();
    // commit_log next
    open_repository
        .expect_commit_log()
        .times(1)
        .with(eq(branches.next()), eq([main_commit.clone()]))
        .return_once(move |_, _| Ok(next_branch_log));
    // commit_log dev
    open_repository
        .expect_commit_log()
        .times(1)
        .with(eq(branches.dev()), eq([main_commit]))
        .return_once(|_, _| Ok(dev_branch_log));

    //when
    let (addr, log) = when::start_actor_with_open_repository(
        Box::new(open_repository),
        repo_details,
        given::a_forge(),
    );
    tell!(addr, ValidateRepo::new(MessageToken::default()))?;

    //then
    log.require_message_containing(format!("send: CheckCIStatus({next_commit:?})"))
        .await?;
    Ok(())
}

#[test_log::test(tokio::test)]
async fn repo_with_dev_and_next_on_main_should_do_nothing() -> TestResult {
    // Do nothing, when the situation changes we will hear about it via a webhook
    //given
    let fs = given::a_filesystem();
    let (mut open_repository, repo_details) = given::an_open_repository(&fs);
    #[allow(clippy::unwrap_used)]
    let repo_config = repo_details.repo_config.clone().unwrap();
    // Validate repo branches
    expect::fetch_ok(&mut open_repository);
    let branches = repo_config.branches();
    // commit_log main
    let main_commit = expect::main_commit_log(&mut open_repository, branches.main());
    // next - on main
    let next_branch_log = vec![main_commit.clone(), given::a_commit()];
    // dev - on next
    let dev_branch_log = next_branch_log.clone();
    // commit_log next
    open_repository
        .expect_commit_log()
        .times(1)
        .with(eq(branches.next()), eq([main_commit.clone()]))
        .return_once(move |_, _| Ok(next_branch_log));
    // commit_log dev
    let dev_commit_log = dev_branch_log.clone();
    open_repository
        .expect_commit_log()
        .times(1)
        .with(eq(branches.dev()), eq([main_commit]))
        .return_once(move |_, _| Ok(dev_commit_log));

    //when
    let (addr, log) = when::start_actor_with_open_repository(
        Box::new(open_repository),
        repo_details,
        given::a_forge(),
    );
    tell!(addr, ValidateRepo::new(MessageToken::default()))?;

    //then
    log.no_message_contains("send:").await?;
    Ok(())
}

#[test_log::test(tokio::test)]
async fn repo_with_dev_ahead_of_next_should_advance_next() -> TestResult {
    //given
    let fs = given::a_filesystem();
    let (mut open_repository, repo_details) = given::an_open_repository(&fs);
    #[allow(clippy::unwrap_used)]
    let repo_config = repo_details.repo_config.clone().unwrap();
    // Validate repo branches
    expect::fetch_ok(&mut open_repository);
    let branches = repo_config.branches();
    // commit_log main
    let main_commit = expect::main_commit_log(&mut open_repository, branches.main());
    // next - on main
    let next_commit = main_commit.clone();
    let next_branch_log = vec![main_commit.clone(), given::a_commit()];
    // dev - ahead of next
    let dev_commit = given::a_named_commit("dev");
    let dev_branch_log = vec![dev_commit, main_commit.clone()];
    // commit_log next
    open_repository
        .expect_commit_log()
        .times(1)
        .with(eq(branches.next()), eq([main_commit.clone()]))
        .return_once(move |_, _| Ok(next_branch_log));
    // commit_log dev
    let dev_commit_log = dev_branch_log.clone();
    open_repository
        .expect_commit_log()
        .times(1)
        .with(eq(branches.dev()), eq([main_commit.clone()]))
        .return_once(move |_, _| Ok(dev_commit_log));

    //when
    let (addr, log) = when::start_actor_with_open_repository(
        Box::new(open_repository),
        repo_details,
        given::a_forge(),
    );
    tell!(addr, ValidateRepo::new(MessageToken::default()))?;

    //then
    let expected = AdvanceNext::new(AdvanceNextPayload {
        next: next_commit,
        main: main_commit,
        dev_commit_history: dev_branch_log,
    });
    log.require_message_containing(format!("send: {expected:?}"))
        .await?;
    Ok(())
}

#[test_log::test(tokio::test)]
async fn repo_with_dev_not_ahead_of_main_should_notify_user() -> TestResult {
    //given
    let fs = given::a_filesystem();
    let (mut open_repository, repo_details) = given::an_open_repository(&fs);
    #[allow(clippy::unwrap_used)]
    let repo_config = repo_details.repo_config.clone().unwrap();
    // Validate repo branches
    expect::fetch_ok(&mut open_repository);
    let branches = repo_config.branches();
    // commit_log main
    let main_commit = expect::main_commit_log(&mut open_repository, branches.main());
    // next - on main
    let next_commit = main_commit.clone();
    let next_branch_log = vec![next_commit.clone(), given::a_commit()];
    // dev - not ahead of next
    let dev_commit = given::a_named_commit("dev");
    let dev_branch_log = vec![dev_commit];
    // commit_log next
    open_repository
        .expect_commit_log()
        .times(1)
        .with(eq(branches.next()), eq([main_commit.clone()]))
        .return_once(move |_, _| Ok(next_branch_log));
    // commit_log dev
    let dev_commit_log = dev_branch_log.clone();
    open_repository
        .expect_commit_log()
        .times(1)
        .with(eq(branches.dev()), eq([main_commit]))
        .return_once(move |_, _| Ok(dev_commit_log));

    //when
    let (addr, log) = when::start_actor_with_open_repository(
        Box::new(open_repository),
        repo_details,
        given::a_forge(),
    );
    tell!(addr, ValidateRepo::new(MessageToken::default()))?;

    //then
    log.require_message_containing("send: NotifyUser").await?;
    Ok(())
}

#[test_log::test(tokio::test)]
async fn should_accept_message_with_current_token() -> TestResult {
    //given
    let fs = given::a_filesystem();
    let repo_details = given::repo_details(&fs);
    let net: Net = given::a_network().into();
    let server_actor_ref = given::a_server_actor(fs.as_real(), net.clone());
    let (actor, log) = given::a_repo_actor(
        repo_details,
        git::repository::factory::mock(),
        given::a_forge(),
        server_actor_ref,
        net.clone(),
    );
    let actor = actor.with_message_token(MessageToken::new(2_u32));
    let addr = kameo::spawn(actor);
    //
    //when
    tell!(addr, ValidateRepo::new(MessageToken::new(2_u32)))?;

    //then
    log.require_message_containing("accepted token: 2").await?;
    net.assert_no_unused_plans();
    Ok(())
}

#[test_log::test(tokio::test)]
async fn should_accept_message_with_new_token() -> TestResult {
    //given
    let fs = given::a_filesystem();
    let repo_details = given::repo_details(&fs);
    let net: Net = given::a_network().into();

    //when
    let (actor, log) = given::a_repo_actor(
        repo_details,
        git::repository::factory::mock(),
        given::a_forge(),
        given::a_server_actor(fs.as_real(), net.clone()),
        net.clone(),
    );
    let actor = actor.with_message_token(MessageToken::new(2_u32));
    let addr = kameo::spawn(actor);
    tell!(addr, ValidateRepo::new(MessageToken::new(3_u32)))?;

    //then
    log.require_message_containing("accepted token: 3").await?;
    net.assert_no_unused_plans();
    Ok(())
}

#[test_log::test(tokio::test)]
async fn should_reject_message_with_expired_token() -> TestResult {
    //given
    let fs = given::a_filesystem();
    let repo_details = given::repo_details(&fs);
    let net: Net = given::a_network().into();

    //when
    let (actor, log) = given::a_repo_actor(
        repo_details,
        git::repository::factory::mock(),
        given::a_forge(),
        given::a_server_actor(fs.as_real(), net.clone()),
        net.clone(),
    );
    let actor = actor.with_message_token(MessageToken::new(4_u32));
    let addr = kameo::spawn(actor);
    tell!(addr, ValidateRepo::new(MessageToken::new(3_u32)))?;

    //then
    log.no_message_contains("accepted token").await?;
    net.assert_no_unused_plans();
    Ok(())
}

#[test_log::test(tokio::test)]
// NOTE: failed then passed on retry: count = 6
async fn should_send_validate_repo_when_retryable_error() -> TestResult {
    //given
    let fs = given::a_filesystem();
    let (mut open_repository, repo_details) = given::an_open_repository(&fs);
    open_repository.expect_fetch().return_once(|| Ok(()));
    open_repository
        .expect_commit_log()
        .return_once(|_, _| Err(git::commit::log::Error::Lock));

    //when
    let (addr, log) = when::start_actor_with_open_repository(
        Box::new(open_repository),
        repo_details,
        given::a_forge(),
    );
    tell!(addr, ValidateRepo::new(MessageToken::default()))?;
    tokio::time::sleep(std::time::Duration::from_millis(2)).await;

    //then
    log.require_message_containing("accepted token: 0").await?;
    log.require_message_containing("send: ValidateRepo").await?;
    Ok(())
}

#[test_log::test(tokio::test)]
async fn should_send_notify_user_when_non_retryable_error() -> TestResult {
    //given
    let fs = given::a_filesystem();
    let (mut open_repository, repo_details) = given::an_open_repository(&fs);
    #[allow(clippy::unwrap_used)]
    let repo_config = repo_details.repo_config.clone().unwrap();
    open_repository.expect_fetch().return_once(|| Ok(()));

    // branches are all unrelated - non-retryable until each branch is updated
    let branches = repo_config.branches();
    // commit_log main
    let main_commit = expect::main_commit_log(&mut open_repository, branches.main());
    // next
    let next_branch_log = vec![given::a_commit()];
    // dev
    let dev_branch_log = vec![given::a_commit()];
    // commit_log next
    open_repository
        .expect_commit_log()
        .times(1)
        .with(eq(branches.next()), eq([main_commit.clone()]))
        .return_once(move |_, _| Ok(next_branch_log));
    // commit_log dev
    open_repository
        .expect_commit_log()
        .times(1)
        .with(eq(branches.dev()), eq([main_commit]))
        .return_once(|_, _| Ok(dev_branch_log));

    //when
    let (addr, log) = when::start_actor_with_open_repository(
        Box::new(open_repository),
        repo_details,
        given::a_forge(),
    );
    tell!(addr, ValidateRepo::new(MessageToken::default()))?;
    tokio::time::sleep(std::time::Duration::from_millis(1)).await;

    //then
    log.require_message_containing("accepted token").await?;
    log.require_message_containing("send: NotifyUser").await?;
    Ok(())
}
