use kxio::net::Net;

use crate::{repo::messages::WebhookNotification, tell};

//
use super::*;

#[tokio::test]
async fn when_no_expected_auth_token_drop_notification() -> TestResult {
    //given
    let fs = given::a_filesystem();
    let repo_details = given::repo_details(&fs);
    let forge_alias = given::a_forge_alias();
    let repo_alias = given::a_repo_alias();
    let headers = BTreeMap::new();
    let body = Body::new(String::new());
    let forge_notification = ForgeNotification::new(forge_alias, repo_alias, headers, body);
    let repository_factory = MockRepositoryFactory::new();
    let net: Net = given::a_network().into();
    let (actor, log) = given::a_repo_actor(
        repo_details,
        Box::new(repository_factory),
        given::a_forge(),
        given::a_server_actor(fs.as_real(), net.clone()),
        net.clone(),
    );
    let actor = actor.with_webhook_auth(None);

    //when
    tell!(
        kameo::spawn(actor),
        WebhookNotification::new(forge_notification)
    )?;

    //then
    log.no_message_contains("send").await?;
    log.require_message_containing("server has no auth token")
        .await?;
    net.assert_no_unused_plans();
    Ok(())
}

#[tokio::test]
async fn when_no_repo_config_drop_notification() -> TestResult {
    //given
    let fs = given::a_filesystem();
    let repo_details = given::repo_details(&fs).with_repo_config(None);
    let forge_alias = given::a_forge_alias();
    let repo_alias = given::a_repo_alias();
    let headers = BTreeMap::new();
    let body = Body::new(String::new());
    let forge_notification = ForgeNotification::new(forge_alias, repo_alias, headers, body);
    let repository_factory = MockRepositoryFactory::new();
    let net: Net = given::a_network().into();
    let (actor, log) = given::a_repo_actor(
        repo_details,
        Box::new(repository_factory),
        given::a_forge(),
        given::a_server_actor(fs.as_real(), net.clone()),
        net.clone(),
    );
    let actor = actor.with_webhook_auth(Some(given::a_webhook_auth()));

    //when
    tell!(
        kameo::spawn(actor),
        WebhookNotification::new(forge_notification)
    )?;

    //then
    log.no_message_contains("send").await?;
    log.require_message_containing("server has no repo config")
        .await?;
    net.assert_no_unused_plans();
    Ok(())
}

#[tokio::test]
async fn when_message_auth_is_invalid_drop_notification() -> TestResult {
    //given
    let fs = given::a_filesystem();
    let repo_details = given::repo_details(&fs);
    let forge_alias = given::a_forge_alias();
    let repo_alias = given::a_repo_alias();
    let headers = BTreeMap::new();
    let body = Body::new(String::new());
    let forge_notification = ForgeNotification::new(forge_alias, repo_alias, headers, body);
    let repository_factory = MockRepositoryFactory::new();
    let mut forge = given::a_forge();
    forge
        .expect_is_message_authorised()
        .return_once(|_, _| false); // is not valid
    let net: Net = given::a_network().into();
    let (actor, log) = given::a_repo_actor(
        repo_details,
        Box::new(repository_factory),
        forge,
        given::a_server_actor(fs.as_real(), net.clone()),
        net.clone(),
    );
    let actor = actor.with_webhook_auth(Some(given::a_webhook_auth()));

    //when
    tell!(
        kameo::spawn(actor),
        WebhookNotification::new(forge_notification)
    )?;

    //then
    log.no_message_contains("send").await?;
    log.require_message_containing("message authorisation is invalid")
        .await?;
    net.assert_no_unused_plans();
    Ok(())
}

#[tokio::test]
async fn when_message_is_ignorable_drop_notification() -> TestResult {
    //given
    let fs = given::a_filesystem();
    let repo_details = given::repo_details(&fs);
    let forge_alias = given::a_forge_alias();
    let repo_alias = given::a_repo_alias();
    let headers = BTreeMap::new();
    let body = Body::new(String::new());
    let forge_notification = ForgeNotification::new(forge_alias, repo_alias, headers, body);
    let repository_factory = MockRepositoryFactory::new();
    let mut forge = given::a_forge();
    forge
        .expect_is_message_authorised()
        .return_once(|_, _| true); // is valid
    forge.expect_should_ignore_message().returning(|_| true);
    forge
        .expect_parse_webhook_body()
        .return_once(|_| Err(git::forge::webhook::Error::NetworkResponseEmpty));
    let net: Net = given::a_network().into();
    let (actor, log) = given::a_repo_actor(
        repo_details,
        Box::new(repository_factory),
        forge,
        given::a_server_actor(fs.as_real(), net.clone()),
        net.clone(),
    );
    let actor = actor.with_webhook_auth(Some(given::a_webhook_auth()));

    //when
    tell!(
        kameo::spawn(actor),
        WebhookNotification::new(forge_notification)
    )?;

    //then
    log.no_message_contains("send").await?;
    log.require_message_containing("forge sent ignorable message")
        .await?;
    net.assert_no_unused_plans();
    Ok(())
}

#[tokio::test]
async fn when_message_is_not_a_push_drop_notification() -> TestResult {
    //given
    let fs = given::a_filesystem();
    let repo_details = given::repo_details(&fs);
    let forge_alias = given::a_forge_alias();
    let repo_alias = given::a_repo_alias();
    let headers = BTreeMap::new();
    let body = Body::new(String::new());
    let forge_notification = ForgeNotification::new(forge_alias, repo_alias, headers, body);
    let repository_factory = MockRepositoryFactory::new();
    let mut forge = given::a_forge();
    forge
        .expect_is_message_authorised()
        .return_once(|_, _| true); // is valid
    forge.expect_should_ignore_message().returning(|_| false);
    forge
        .expect_parse_webhook_body()
        .return_once(|_| Err(git::forge::webhook::Error::NetworkResponseEmpty));
    let net: Net = given::a_network().into();
    let (actor, log) = given::a_repo_actor(
        repo_details,
        Box::new(repository_factory),
        forge,
        given::a_server_actor(fs.as_real(), net.clone()),
        net.clone(),
    );
    let actor = actor.with_webhook_auth(Some(given::a_webhook_auth()));

    //when
    tell!(
        kameo::spawn(actor),
        WebhookNotification::new(forge_notification)
    )?;

    //then
    log.no_message_contains("send").await?;
    log.require_message_containing("message parse error - not a push")
        .await?;
    net.assert_no_unused_plans();
    Ok(())
}

#[tokio::test]
async fn when_message_is_push_on_unknown_branch_drop_notification() -> TestResult {
    //given
    let fs = given::a_filesystem();
    let repo_config = given::a_repo_config();
    let repo_details = given::repo_details(&fs).with_repo_config(Some(repo_config.clone()));
    let forge_alias = given::a_forge_alias();
    let repo_alias = given::a_repo_alias();
    let headers = BTreeMap::new();
    let body = Body::new(String::new());
    let forge_notification = ForgeNotification::new(forge_alias, repo_alias, headers, body);
    let repository_factory = MockRepositoryFactory::new();
    let commit = given::a_commit();
    let push = given::a_push()
        .with_branch(given::a_branch_name("unknown"))
        .with_sha(commit.sha().to_string())
        .with_message(commit.message().to_string());
    let mut forge = given::a_forge();
    forge
        .expect_is_message_authorised()
        .return_once(|_, _| true); // is valid
    forge.expect_should_ignore_message().returning(|_| false);
    forge.expect_parse_webhook_body().return_once(|_| Ok(push));
    let net: Net = given::a_network().into();
    let (actor, log) = given::a_repo_actor(
        repo_details,
        Box::new(repository_factory),
        forge,
        given::a_server_actor(fs.as_real(), net.clone()),
        net.clone(),
    );
    let actor = actor
        .with_webhook_auth(Some(given::a_webhook_auth()))
        .with_last_main_commit(commit);

    //when
    tell!(
        kameo::spawn(actor),
        WebhookNotification::new(forge_notification)
    )?;

    //then
    log.no_message_contains("send").await?;
    log.require_message_containing("unknown branch").await?;
    net.assert_no_unused_plans();
    Ok(())
}

#[tokio::test]
async fn when_message_is_push_already_seen_commit_to_main() -> TestResult {
    //given
    let fs = given::a_filesystem();
    let repo_config = given::a_repo_config();
    let repo_details = given::repo_details(&fs).with_repo_config(Some(repo_config.clone()));
    let forge_alias = given::a_forge_alias();
    let repo_alias = given::a_repo_alias();
    let headers = BTreeMap::new();
    let body = Body::new(String::new());
    let forge_notification = ForgeNotification::new(forge_alias, repo_alias, headers, body);
    let repository_factory = MockRepositoryFactory::new();
    let commit = given::a_commit();
    let main = repo_config.branches().main();
    let push = given::a_push()
        .with_branch(main.clone())
        .with_sha(commit.sha().to_string())
        .with_message(commit.message().to_string());
    let mut forge = given::a_forge();
    forge
        .expect_is_message_authorised()
        .return_once(|_, _| true); // is valid
    forge.expect_should_ignore_message().returning(|_| false);
    forge.expect_parse_webhook_body().return_once(|_| Ok(push));
    let net: Net = given::a_network().into();
    let (actor, log) = given::a_repo_actor(
        repo_details,
        Box::new(repository_factory),
        forge,
        given::a_server_actor(fs.as_real(), net.clone()),
        net.clone(),
    );
    let actor = actor
        .with_webhook_auth(Some(given::a_webhook_auth()))
        .with_last_main_commit(commit);

    //when
    tell!(
        kameo::spawn(actor),
        WebhookNotification::new(forge_notification)
    )?;

    //then
    log.no_message_contains("send").await?;
    log.require_message_containing(format!("not a new commit on {main}"))
        .await?;
    net.assert_no_unused_plans();
    Ok(())
}

#[tokio::test]
async fn when_message_is_push_already_seen_commit_to_next() -> TestResult {
    //given
    let fs = given::a_filesystem();
    let repo_config = given::a_repo_config();
    let repo_details = given::repo_details(&fs).with_repo_config(Some(repo_config.clone()));
    let forge_alias = given::a_forge_alias();
    let repo_alias = given::a_repo_alias();
    let headers = BTreeMap::new();
    let body = Body::new(String::new());
    let forge_notification = ForgeNotification::new(forge_alias, repo_alias, headers, body);
    let repository_factory = MockRepositoryFactory::new();
    let commit = given::a_commit();
    let next_branch = repo_config.branches().next();
    let push = given::a_push()
        .with_branch(next_branch.clone())
        .with_sha(commit.sha().to_string())
        .with_message(commit.message().to_string());
    let mut forge = given::a_forge();
    forge
        .expect_is_message_authorised()
        .return_once(|_, _| true); // is valid
    forge.expect_should_ignore_message().returning(|_| false);
    forge.expect_parse_webhook_body().return_once(|_| Ok(push));
    let net: Net = given::a_network().into();
    let (actor, log) = given::a_repo_actor(
        repo_details,
        Box::new(repository_factory),
        forge,
        given::a_server_actor(fs.as_real(), net.clone()),
        net.clone(),
    );
    let actor = actor
        .with_webhook_auth(Some(given::a_webhook_auth()))
        .with_last_next_commit(commit);

    //when
    tell!(
        kameo::spawn(actor),
        WebhookNotification::new(forge_notification)
    )?;

    //then
    log.no_message_contains("send").await?;
    log.require_message_containing(format!("not a new commit on {next_branch}"))
        .await?;
    net.assert_no_unused_plans();
    Ok(())
}

#[tokio::test]
async fn when_message_is_push_already_seen_commit_to_dev() -> TestResult {
    //given
    let fs = given::a_filesystem();
    let repo_config = given::a_repo_config();
    let repo_details = given::repo_details(&fs).with_repo_config(Some(repo_config.clone()));
    let forge_alias = given::a_forge_alias();
    let repo_alias = given::a_repo_alias();
    let headers = BTreeMap::new();
    let body = Body::new(String::new());
    let forge_notification = ForgeNotification::new(forge_alias, repo_alias, headers, body);
    let repository_factory = MockRepositoryFactory::new();
    let commit = given::a_commit();
    let dev = repo_config.branches().dev();
    let push = given::a_push()
        .with_branch(dev.clone())
        .with_sha(commit.sha().to_string())
        .with_message(commit.message().to_string());
    let mut forge = given::a_forge();
    forge
        .expect_is_message_authorised()
        .return_once(|_, _| true); // is valid
    forge.expect_should_ignore_message().returning(|_| false);
    forge.expect_parse_webhook_body().return_once(|_| Ok(push));
    let net: Net = given::a_network().into();
    let (actor, log) = given::a_repo_actor(
        repo_details,
        Box::new(repository_factory),
        forge,
        given::a_server_actor(fs.as_real(), net.clone()),
        net.clone(),
    );
    let actor = actor
        .with_webhook_auth(Some(given::a_webhook_auth()))
        .with_last_dev_commit(commit);

    //when
    tell!(
        kameo::spawn(actor),
        WebhookNotification::new(forge_notification)
    )?;

    //then
    log.no_message_contains("send").await?;
    log.require_message_containing(format!("not a new commit on {dev}"))
        .await?;
    net.assert_no_unused_plans();
    Ok(())
}

#[tokio::test]
async fn when_message_is_push_new_commit_to_main_should_stash_and_validate_repo() -> TestResult {
    //given
    let fs = given::a_filesystem();
    let repo_config = given::a_repo_config();
    let repo_details = given::repo_details(&fs).with_repo_config(Some(repo_config.clone()));
    let forge_alias = given::a_forge_alias();
    let repo_alias = given::a_repo_alias();
    let headers = BTreeMap::new();
    let body = Body::new(String::new());
    let forge_notification = ForgeNotification::new(forge_alias, repo_alias, headers, body);
    let repository_factory = MockRepositoryFactory::new();
    let push_commit = given::a_commit();
    let push = given::a_push()
        .with_branch(repo_config.branches().main())
        .with_sha(push_commit.sha().to_string())
        .with_message(push_commit.message().to_string());
    let mut forge = given::a_forge();
    forge
        .expect_is_message_authorised()
        .return_once(|_, _| true); // is valid
    forge.expect_should_ignore_message().returning(|_| false);
    forge.expect_parse_webhook_body().return_once(|_| Ok(push));
    let net: Net = given::a_network().into();
    let (actor, log) = given::a_repo_actor(
        repo_details,
        Box::new(repository_factory),
        forge,
        given::a_server_actor(fs.as_real(), net.clone()),
        net.clone(),
    );
    let actor = actor
        .with_webhook_auth(Some(given::a_webhook_auth()))
        .with_last_main_commit(given::a_commit());

    //when
    let addr = kameo::spawn(actor);
    tell!(addr, WebhookNotification::new(forge_notification))?;

    //then
    let view = addr
        .ask(ExamineActor)
        .await
        .map_err(|e| format!("examine actor: {e:?}"))?;
    assert_eq!(view.last_main_commit, Some(push_commit));
    log.require_message_containing("send: ValidateRepo").await?;
    net.assert_no_unused_plans();
    Ok(())
}

#[tokio::test]
async fn when_message_is_push_new_commit_to_next_should_stash_and_validate_repo() -> TestResult {
    //given
    let fs = given::a_filesystem();
    let repo_config = given::a_repo_config();
    let repo_details = given::repo_details(&fs).with_repo_config(Some(repo_config.clone()));
    let forge_alias = given::a_forge_alias();
    let repo_alias = given::a_repo_alias();
    let headers = BTreeMap::new();
    let body = Body::new(String::new());
    let forge_notification = ForgeNotification::new(forge_alias, repo_alias, headers, body);
    let repository_factory = MockRepositoryFactory::new();
    let push_commit = given::a_commit();
    let push = given::a_push()
        .with_branch(repo_config.branches().next())
        .with_sha(push_commit.sha().to_string())
        .with_message(push_commit.message().to_string());
    let mut forge = given::a_forge();
    forge
        .expect_is_message_authorised()
        .return_once(|_, _| true); // is valid
    forge.expect_should_ignore_message().returning(|_| false);
    forge.expect_parse_webhook_body().return_once(|_| Ok(push));
    let net: Net = given::a_network().into();
    let (actor, log) = given::a_repo_actor(
        repo_details,
        Box::new(repository_factory),
        forge,
        given::a_server_actor(fs.as_real(), net.clone()),
        net.clone(),
    );
    let actor = actor
        .with_webhook_auth(Some(given::a_webhook_auth()))
        .with_last_next_commit(given::a_commit());

    //when
    let addr = kameo::spawn(actor);
    tell!(addr, WebhookNotification::new(forge_notification))?;

    //then
    let view = addr
        .ask(ExamineActor)
        .await
        .map_err(|e| format!("examine actor: {e:?}"))?;
    assert_eq!(view.last_next_commit, Some(push_commit));
    log.require_message_containing("send: ValidateRepo").await?;
    net.assert_no_unused_plans();
    Ok(())
}

#[tokio::test]
async fn when_message_is_push_new_commit_to_dev_should_stash_and_validate_repo() -> TestResult {
    //given
    let fs = given::a_filesystem();
    let repo_config = given::a_repo_config();
    let repo_details = given::repo_details(&fs).with_repo_config(Some(repo_config.clone()));
    let forge_alias = given::a_forge_alias();
    let repo_alias = given::a_repo_alias();
    let headers = BTreeMap::new();
    let body = Body::new(String::new());
    let forge_notification = ForgeNotification::new(forge_alias, repo_alias, headers, body);
    let repository_factory = MockRepositoryFactory::new();
    let push_commit = given::a_commit();
    let push = given::a_push()
        .with_branch(repo_config.branches().dev())
        .with_sha(push_commit.sha().to_string())
        .with_message(push_commit.message().to_string());
    let mut forge = given::a_forge();
    forge
        .expect_is_message_authorised()
        .return_once(|_, _| true); // is valid
    forge.expect_should_ignore_message().returning(|_| false);
    forge.expect_parse_webhook_body().return_once(|_| Ok(push));
    let net: Net = given::a_network().into();
    let (actor, log) = given::a_repo_actor(
        repo_details,
        Box::new(repository_factory),
        forge,
        given::a_server_actor(fs.as_real(), net.clone()),
        net.clone(),
    );
    let actor = actor
        .with_webhook_auth(Some(given::a_webhook_auth()))
        .with_last_dev_commit(given::a_commit());

    //when
    let addr = kameo::spawn(actor);
    tell!(addr, WebhookNotification::new(forge_notification))?;

    //then
    let view = addr
        .ask(ExamineActor)
        .await
        .map_err(|e| format!("examine actor: {e:?}"))?;
    assert_eq!(view.last_dev_commit, Some(push_commit));
    log.require_message_containing("send: ValidateRepo").await?;
    net.assert_no_unused_plans();
    Ok(())
}
