//
use std::{
    collections::BTreeMap,
    sync::{Arc, RwLock},
};

use kameo::actor::pubsub::PubSub;

use git_next_core::{
    git,
    server::{AppConfig, Http, Listen, ListenUrl, Shout, Storage},
};

use crate::{
    server::actor::{tests::given, ReceiveAppConfig, ServerActor},
    tell,
};

type TestResult = Result<(), Box<dyn std::error::Error>>;

#[test_log::test(tokio::test)]
async fn when_webhook_url_has_trailing_slash_should_not_send() -> TestResult {
    //given
    // parameters
    let fs = given::a_filesystem();
    let mock_net = given::a_network();
    let alerts = given::an_alerts_actor(mock_net.clone().into());
    let repo = git::repository::factory::mock();
    let duration = std::time::Duration::from_millis(1);

    let file_update_subs = kameo::spawn(PubSub::new());

    // sut
    let server = ServerActor::new(
        false, // ui
        fs.as_real(),
        mock_net.clone().into(),
        alerts,
        file_update_subs,
        repo,
        duration,
    );

    // collaborators
    let listen = Listen::new(
        Http::new("0.0.0.0".to_string(), 80),
        ListenUrl::new("http://localhost/".to_string()), // with trailing slash
    );
    let shout = Shout::default();
    let server_storage = Storage::new((fs.base()).to_path_buf());
    let repos = BTreeMap::default();

    // debugging
    let message_log: Arc<RwLock<Vec<String>>> = Arc::new(RwLock::new(vec![]));
    let server = server.with_message_log(Some(message_log.clone()));

    //when
    tell!(
        kameo::spawn(server),
        ReceiveAppConfig::new(AppConfig::new(listen, shout, server_storage, repos,))
    )?;
    tokio::time::sleep(std::time::Duration::from_millis(1)).await;

    //then
    // INFO: assert that ReceiveValidServerConfig is NOT sent
    tracing::debug!(?message_log, "");
    assert!(message_log.read().iter().any(|log| !log
        .iter()
        .any(|line| line == "send: ReceiveValidServerConfig")));
    mock_net.assert_no_unused_plans();

    Ok(())
}
