//
use kameo::message::{Context, Message};
use ratatui::style::Color;
use tracing::debug;

use crate::{
    server::actor::messages::{RepoUpdate, ServerUpdate},
    tui::{actor::ServerState, Tui},
};

static OKAY: Color = Color::Green;
static PREP: Color = Color::Gray;
static ACTING: Color = Color::LightBlue;
static WARN: Color = Color::Red;

impl Message<ServerUpdate> for Tui {
    type Reply = ();

    async fn handle(
        &mut self,
        msg: ServerUpdate,
        _ctx: Context<'_, Self, Self::Reply>,
    ) -> Self::Reply {
        self.state.tap();
        match msg {
            ServerUpdate::AppConfigLoaded { app_config } => {
                self.state.mode = ServerState::from(app_config);
            }

            ServerUpdate::RepoUpdate {
                forge_alias,
                repo_alias,
                repo_update,
            } => {
                if let ServerState::Configured { forges } = &mut self.state.mode {
                    let Some(forge_state) = forges.get_mut(&forge_alias) else {
                        debug!("ServerState::Configured: no forge state available");
                        return;
                    };
                    let Some(repo_state) = forge_state.repos.get_mut(&repo_alias) else {
                        debug!("ServerState::Configured: no repo state available");
                        return;
                    };
                    repo_state.clear_alert();
                    debug!(?repo_update, "ServerState::Configured: RepoUpdate");
                    handle_repo_update(repo_update, repo_state);
                }
            }
        }
    }
}

fn handle_repo_update(repo_update: RepoUpdate, repo_state: &mut crate::tui::actor::RepoState) {
    match repo_update {
        RepoUpdate::Branches { branches } => {
            repo_state.update_branches(branches);
        }
        RepoUpdate::Log { log } => {
            repo_state.update_log(log);
        }
        RepoUpdate::ValidateRepo => {
            repo_state.update_message("polling...", ACTING);
        }
        RepoUpdate::Okay { main, next, dev } => {
            repo_state.clear_alert();
            repo_state.update_message("okay", OKAY);
            *repo_state = repo_state.clone().ready(main, next, dev);
        }
        RepoUpdate::Alert { alert } => {
            repo_state.alert(alert);
        }
        RepoUpdate::CheckingCI => {
            repo_state.update_message("Checking CI status", ACTING);
        }
        RepoUpdate::AdvancingNext { commit, force: _ } => {
            repo_state.update_message(format!("advancing next to {commit}"), ACTING);
        }
        RepoUpdate::NextUpdated => {
            repo_state.update_message("next updated - pause while CI starts", OKAY);
        }
        RepoUpdate::AdvancingMain { commit } => {
            repo_state.update_message(format!("advancing main to {commit}"), ACTING);
        }
        RepoUpdate::MainUpdated => {
            repo_state.update_message("main updated", OKAY);
        }
        RepoUpdate::Opening => {
            repo_state.update_message("opening...", PREP);
        }
        RepoUpdate::Opened => {
            repo_state.update_message("opened", PREP);
        }
        RepoUpdate::LoadingConfigFromRepo => {
            repo_state.update_message("loading config from repo...", PREP);
        }
        RepoUpdate::ReceiveCIStatus { status } => {
            repo_state.update_message(format!("ci status: {status:?}"), WARN);
        }
        RepoUpdate::ReceiveRepoConfig { repo_config: _ } => {
            repo_state.update_message("loaded config from repo", PREP);
        }
        RepoUpdate::RegisteringWebhook => {
            repo_state.update_message("registering webhook...", PREP);
        }
        RepoUpdate::UnregisteringWebhook => {
            repo_state.update_message("unregistering webhook...", PREP);
        }
        RepoUpdate::WebhookReceived { branch, push: _ } => {
            repo_state.update_message(format!("webhook update: {branch:?}"), ACTING);
        }
        RepoUpdate::RegisteredWebhook => {
            repo_state.update_message("registered webhook", PREP);
        }
    }
}
