//
#![allow(clippy::expect_used)]

use crate::{Github, GithubState, GithubStatus};
use git_next_core::{
    git::{self, forge::commit::Status, ForgeLike},
    server::ListenUrl,
    webhook::{self, forge_notification::Body},
    ForgeAlias, ForgeConfig, ForgeNotification, ForgeType, GitDir, Hostname, RepoAlias,
    RepoBranches, RepoPath, ServerRepoConfig, StoragePathType, WebhookAuth, WebhookId,
};

use assert2::let_assert;
use kxio::net::{MockNet, StatusCode};
use rand::RngCore;
use serde::Serialize;
use serde_json::json;

use std::{collections::BTreeMap, path::PathBuf};

mod github {
    use super::*;

    #[test]
    fn should_return_name() {
        let mock_net = given::a_network();
        let forge = given::a_github_forge(&given::repo_details(), mock_net.clone());
        assert_eq!(forge.name(), "github");
        mock_net.assert_no_unused_plans();
    }

    mod is_message_authorised {

        use super::*;

        #[test]
        fn should_return_true_with_valid_header() {
            let mock_net = given::a_network();
            let forge = given::a_github_forge(&given::repo_details(), mock_net.clone());
            let auth = given::a_webhook_auth();
            let message = given::a_webhook_message(given::Header::Valid(
                auth.clone(),
                given::a_webhook_message_body(),
            ));
            assert!(forge.is_message_authorised(&message, &auth));
            mock_net.assert_no_unused_plans();
        }
        #[test]
        fn should_return_false_with_missing_header() {
            let mock_net = given::a_network();
            let forge = given::a_github_forge(&given::repo_details(), mock_net.clone());
            let auth = given::a_webhook_auth();
            let message = given::a_webhook_message(given::Header::Missing);
            assert!(!forge.is_message_authorised(&message, &auth));
            mock_net.assert_no_unused_plans();
        }
        #[test]
        fn should_return_false_with_invalid_header() {
            let mock_net = given::a_network();
            let forge = given::a_github_forge(&given::repo_details(), mock_net.clone());
            let auth = given::a_webhook_auth();
            let message = given::a_webhook_message(given::Header::Invalid);
            assert!(!forge.is_message_authorised(&message, &auth));
            mock_net.assert_no_unused_plans();
        }
    }

    mod parse_webhook_body {

        use super::*;

        #[test]
        fn should_parse_valid_body() {
            let mock_net = given::a_network();
            let forge = given::a_github_forge(&given::repo_details(), mock_net.clone());
            let repo_branches = given::repo_branches();
            let next = repo_branches.next();
            let sha = given::a_name();
            let message = given::a_name();
            let body = Body::new(
                json!({"ref":format!("refs/heads/{next}"),"after":sha,"head_commit":{"message":message}})
                    .to_string(),
            );
            let_assert!(Ok(push) = forge.parse_webhook_body(&body));
            assert_eq!(push.sha(), sha);
            assert_eq!(push.message(), message);
            assert_eq!(
                push.branch(&repo_branches),
                Some(webhook::push::Branch::Next)
            );
            mock_net.assert_no_unused_plans();
        }

        #[test]
        fn should_error_invalid_body() {
            let mock_net = given::a_network();
            let forge = given::a_github_forge(&given::repo_details(), mock_net.clone());
            let body = Body::new(r#"{"type":"invalid"}"#.to_string());
            let_assert!(Err(_) = forge.parse_webhook_body(&body));
            mock_net.assert_no_unused_plans();
        }
    }

    mod commit_status {

        use super::*;

        macro_rules! status_tests{
            ($($name:ident: $value:expr, )*) => {
                $(
                #[tokio::test]
                async fn $name() {
                    let (states, expected) = $value;
                    let repo_details = given::repo_details();
                    let commit = given::a_commit();
                    let mock_net = given::a_network();
                    given::commit_states(&states, &mock_net, &repo_details, &commit);
                    let forge = given::a_github_forge(&repo_details, mock_net.clone());
                    assert_eq!(forge.commit_status(&commit).await.expect("status"), expected);
                    mock_net.assert_no_unused_plans();
                }
            )*
            }
        }

        status_tests!(
            pass_when_success: ([GithubState::Success], Status::Pass),
            pending_when_pending: ([GithubState::Pending], Status::Pending),
            fail_when_failure: ([GithubState::Failure], Status::Fail),
            fail_when_error: ([GithubState::Error], Status::Fail),
            pending_when_blank: ([GithubState::Blank], Status::Pending),
            pass_wneh_no_checks: ([], Status::Pass),
            pass_when_all_success: ([GithubState::Success, GithubState::Success], Status::Pass),
            fail_when_only_alpha_fails: ([GithubState::Failure, GithubState::Success], Status::Fail),
            fail_when_only_beta_fails: ([GithubState::Success, GithubState::Failure], Status::Fail),
            pending_when_all_pending: ([GithubState::Pending, GithubState::Pending], Status::Pending),
            pending_when_only_alpha_pending: ([GithubState::Pending, GithubState::Success], Status::Pending),
            pending_when_only_beta_pending: ([GithubState::Success, GithubState::Pending], Status::Pending),

        );

        #[tokio::test]
        async fn should_return_pass_for_no_statuses() {
            let repo_details = given::repo_details();
            let commit = given::a_commit();
            let mock_net = given::a_network();
            mock_net
                .on()
                .get(given::a_commit_status_url(&repo_details, &commit))
                .respond(StatusCode::OK)
                .body(json!([]).to_string())
                .expect("mock");
            let forge = given::a_github_forge(&repo_details, mock_net.clone());
            assert_eq!(
                forge.commit_status(&commit).await.expect("status"),
                Status::Pass // no CI checks configured
            );
            mock_net.assert_no_unused_plans();
        }

        #[tokio::test]
        async fn should_return_pending_for_network_error() {
            let repo_details = given::repo_details();
            let commit = given::a_commit();
            let mock_net = given::a_network();
            mock_net
                .on()
                .get(given::a_commit_status_url(&repo_details, &commit))
                .respond(StatusCode::INTERNAL_SERVER_ERROR)
                .body("boom today")
                .expect("mock");
            let forge = given::a_github_forge(&repo_details, mock_net.clone());
            assert_eq!(
                forge.commit_status(&commit).await.expect("status"),
                Status::Pending
            );
            mock_net.assert_no_unused_plans();
        }
    }

    mod list_webhooks {

        use super::*;

        #[tokio::test]
        async fn should_return_a_list_of_matching_webhooks() {
            let repo_details = given::repo_details();
            let repo_listen_url = given::a_repo_listen_url(&repo_details);
            let hostname = repo_details.forge.hostname();
            let repo_path = &repo_details.repo_path;
            let hook_id_1 = given::a_github_webhook_id();
            let hook_id_2 = given::a_github_webhook_id();
            let hook_id_3 = given::a_github_webhook_id();
            let mock_net = given::a_network();
            let args = with::WebhookArgs {
                mock_net: &mock_net,
                hostname,
                repo_path,
            };
            // page 1 with three items
            with::get_webhooks_by_page(
                1,
                &[
                    with::ReturnedWebhook::new(hook_id_1, &repo_listen_url),
                    with::ReturnedWebhook::new(hook_id_2, &repo_listen_url),
                    with::ReturnedWebhook::new(hook_id_3, &given::any_webhook_url()),
                ],
                &args,
            );
            // page 2 with no items - stops pagination
            with::get_webhooks_by_page(2, &[], &args);

            let forge = given::a_github_forge(&repo_details, mock_net.clone());

            let_assert!(Ok(result) = forge.list_webhooks(&repo_listen_url).await);
            assert_eq!(
                result,
                vec![
                    WebhookId::new(format!("{hook_id_1}")),
                    WebhookId::new(format!("{hook_id_2}"))
                ]
            );
            mock_net.assert_no_unused_plans();
        }

        #[tokio::test]
        async fn should_return_any_network_error() {
            let repo_details = given::repo_details();
            let repo_listen_url = given::a_repo_listen_url(&repo_details);
            let hostname = repo_details.forge.hostname();
            let repo_path = &repo_details.repo_path;
            let mock_net = given::a_network();
            mock_net
                .on()
                .get(format!(
                    "https://api.{hostname}/repos/{repo_path}/hooks?page=1"
                ))
                .respond(StatusCode::INTERNAL_SERVER_ERROR)
                .body("error_message")
                .expect("mock");

            let forge = given::a_github_forge(&repo_details, mock_net.clone());

            let_assert!(Err(_) = forge.list_webhooks(&repo_listen_url).await);
            mock_net.assert_no_unused_plans();
        }
    }

    mod unregister_webhook {
        use super::*;

        #[tokio::test]
        async fn should_delete_webhook() {
            let repo_details = given::repo_details();
            let hostname = repo_details.forge.hostname();
            let repo_path = &repo_details.repo_path;
            let webhook_id = given::a_webhook_id();
            let mock_net = given::a_network();
            mock_net
                .on()
                .delete(format!(
                    "https://api.{hostname}/repos/{repo_path}/hooks/{webhook_id}"
                ))
                .respond(StatusCode::OK)
                .body("")
                .expect("mock");

            let forge = given::a_github_forge(&repo_details, mock_net.clone());

            let_assert!(Ok(()) = forge.unregister_webhook(&webhook_id).await);
            mock_net.assert_no_unused_plans();
        }
        #[tokio::test]
        async fn should_return_error_on_network_error() {
            let repo_details = given::repo_details();
            assert!(
                repo_details.repo_config.is_some(),
                "repo_details needs to have repo_config for this test"
            );
            let hostname = repo_details.forge.hostname();
            let repo_path = &repo_details.repo_path;
            let mock_net = given::a_network();
            // unregister the webhook will return empty response
            let webhook_id = given::a_webhook_id();
            mock_net
                .on()
                .delete(format!(
                    "https://api.{hostname}/repos/{repo_path}/hooks/{webhook_id}"
                ))
                .respond(StatusCode::INTERNAL_SERVER_ERROR)
                .mock()
                .expect("mock");

            let forge = given::a_github_forge(&repo_details, mock_net.clone());

            let_assert!(Err(err) = forge.unregister_webhook(&webhook_id).await);
            assert!(
                matches!(err, git::forge::webhook::Error::FailedToRegister(_)),
                "{err:?}"
            );
            mock_net.assert_no_unused_plans();
        }
    }

    mod register_webhook {

        use super::*;

        #[tokio::test]
        async fn should_register_a_new_webhook() {
            let repo_details = given::repo_details();
            assert!(
                repo_details.repo_config.is_some(),
                "repo_details needs to have repo_config for this test"
            );
            let repo_listen_url = given::a_repo_listen_url(&repo_details);
            let hostname = repo_details.forge.hostname();
            let repo_path = &repo_details.repo_path;
            let mock_net = given::a_network();
            let args = with::WebhookArgs {
                mock_net: &mock_net,
                hostname,
                repo_path,
            };
            // there are no existing matching webhooks
            with::get_webhooks_by_page(1, &[], &args);
            // register the webhook will succeed
            let webhook_id = given::a_github_webhook_id();
            mock_net
                .on()
                .post(format!("https://api.{hostname}/repos/{repo_path}/hooks"))
                .respond(StatusCode::OK)
                .body(
                    json!({"id": webhook_id, "config":{"url": repo_listen_url.to_string()}})
                        .to_string(),
                )
                .expect("mock");

            let forge = given::a_github_forge(&repo_details, mock_net.clone());

            let_assert!(Ok(registered_webhook) = forge.register_webhook(&repo_listen_url).await);
            assert_eq!(
                registered_webhook.id(),
                &WebhookId::new(format!("{webhook_id}"))
            );
            mock_net.assert_no_unused_plans();
        }

        #[tokio::test]
        async fn should_abort_if_repo_config_missing() {
            let mut repo_details = given::repo_details();
            repo_details.repo_config.take();
            assert!(
                repo_details.repo_config.is_none(),
                "repo_details needs to NOT have repo_config for this test"
            );
            let repo_listen_url = given::a_repo_listen_url(&repo_details);
            let mock_net = given::a_network();
            let forge = given::a_github_forge(&repo_details, mock_net.clone());
            let_assert!(Err(err) = forge.register_webhook(&repo_listen_url).await);
            assert!(
                matches!(err, git::forge::webhook::Error::NoRepoConfig),
                "{err:?}"
            );
            mock_net.assert_no_unused_plans();
        }

        #[tokio::test]
        async fn should_unregister_existing_webhooks_before_registering() {
            let repo_details = given::repo_details();
            assert!(
                repo_details.repo_config.is_some(),
                "repo_details needs to have repo_config for this test"
            );
            let repo_listen_url = given::a_repo_listen_url(&repo_details);
            let hostname = repo_details.forge.hostname();
            let repo_path = &repo_details.repo_path;
            let mock_net = given::a_network();
            let args = with::WebhookArgs {
                mock_net: &mock_net,
                hostname,
                repo_path,
            };
            let hook_1 = with::ReturnedWebhook::new(given::a_github_webhook_id(), &repo_listen_url);
            let hook_2 = with::ReturnedWebhook::new(given::a_github_webhook_id(), &repo_listen_url);
            let hook_3 =
                with::ReturnedWebhook::new(given::a_github_webhook_id(), &given::any_webhook_url());
            let hooks = [hook_1, hook_2, hook_3];
            // there are three existing webhooks, two are matching webhooks
            with::get_webhooks_by_page(1, &hooks, &args);
            with::get_webhooks_by_page(2, &[], &args);
            // should unregister 1 and 2, but not 3
            with::unregister_webhook(&hooks[0], &args);
            with::unregister_webhook(&hooks[1], &args);
            // register the webhook will succeed
            let webhook_id = given::a_github_webhook_id();
            mock_net
                .on()
                .post(format!("https://api.{hostname}/repos/{repo_path}/hooks"))
                .respond(StatusCode::OK)
                .body(
                    json!({"id": webhook_id, "config":{"url": repo_listen_url.to_string()}})
                        .to_string(),
                )
                .expect("mock");

            let forge = given::a_github_forge(&repo_details, mock_net.clone());

            let_assert!(Ok(registered_webhook) = forge.register_webhook(&repo_listen_url).await);
            assert_eq!(
                registered_webhook.id(),
                &WebhookId::new(format!("{webhook_id}"))
            );
            mock_net.assert_no_unused_plans();
        }

        #[tokio::test]
        async fn should_return_error_if_empty_network_response() {
            let repo_details = given::repo_details();
            assert!(
                repo_details.repo_config.is_some(),
                "repo_details needs to have repo_config for this test"
            );
            let repo_listen_url = given::a_repo_listen_url(&repo_details);
            let hostname = repo_details.forge.hostname();
            let repo_path = &repo_details.repo_path;
            let mock_net = given::a_network();
            let args = with::WebhookArgs {
                mock_net: &mock_net,
                hostname,
                repo_path,
            };
            // there are no existing matching webhooks
            with::get_webhooks_by_page(1, &[], &args);
            // register the webhook will return empty response
            mock_net
                .on()
                .post(format!("https://api.{hostname}/repos/{repo_path}/hooks"))
                .respond(StatusCode::OK)
                .body(json!({}).to_string())
                .expect("mock");

            let forge = given::a_github_forge(&repo_details, mock_net.clone());

            let_assert!(Err(err) = forge.register_webhook(&repo_listen_url).await);
            assert!(
                matches!(err, git::forge::webhook::Error::NetworkResponseEmpty),
                "{err:?}"
            );
            mock_net.assert_no_unused_plans();
        }

        #[tokio::test]
        async fn should_return_error_on_network_error() {
            let repo_details = given::repo_details();
            assert!(
                repo_details.repo_config.is_some(),
                "repo_details needs to have repo_config for this test"
            );
            let repo_listen_url = given::a_repo_listen_url(&repo_details);
            let hostname = repo_details.forge.hostname();
            let repo_path = &repo_details.repo_path;
            let mock_net = given::a_network();
            let args = with::WebhookArgs {
                mock_net: &mock_net,
                hostname,
                repo_path,
            };
            // there are no existing matching webhooks
            with::get_webhooks_by_page(1, &[], &args);
            // register the webhook will return empty response
            mock_net
                .on()
                .post(format!("https://api.{hostname}/repos/{repo_path}/hooks"))
                .respond(StatusCode::INTERNAL_SERVER_ERROR)
                .body("error")
                .expect("mock");

            let forge = given::a_github_forge(&repo_details, mock_net.clone());

            let_assert!(Err(err) = forge.register_webhook(&repo_listen_url).await);
            assert!(
                matches!(err, git::forge::webhook::Error::FailedToRegister(_)),
                "{err:?}"
            );
            mock_net.assert_no_unused_plans();
        }
    }

    pub mod with {
        use git_next_core::server::RepoListenUrl;

        use super::*;

        pub fn get_webhooks_by_page(page: u8, response: &[ReturnedWebhook], args: &WebhookArgs) {
            let hostname = args.hostname;
            let repo_path = args.repo_path;
            args.mock_net
                .on()
                .get(format!(
                    "https://api.{hostname}/repos/{repo_path}/hooks?page={page}"
                ))
                .respond(StatusCode::OK)
                .body(json!(response).to_string())
                .expect("mock");
        }

        pub fn unregister_webhook(hook1: &ReturnedWebhook, args: &WebhookArgs) {
            let webhook_id = hook1.id;
            let hostname = args.hostname;
            let repo_path = args.repo_path;
            args.mock_net
                .on()
                .delete(format!(
                    "https://api.{hostname}/repos/{repo_path}/hooks/{webhook_id}"
                ))
                .respond(StatusCode::OK)
                .body("")
                .expect("mock");
        }
        pub struct WebhookArgs<'a> {
            pub mock_net: &'a kxio::net::MockNet,
            pub hostname: &'a Hostname,
            pub repo_path: &'a RepoPath,
        }
        #[derive(Debug, Serialize)]
        pub struct ReturnedWebhook {
            pub id: i64,
            pub config: Config,
        }
        impl ReturnedWebhook {
            pub fn new(id: i64, url: &RepoListenUrl) -> Self {
                Self {
                    id,
                    config: Config {
                        url: url.to_string(),
                    },
                }
            }
        }
        #[derive(Debug, Serialize)]
        pub struct Config {
            pub url: String,
        }
    }

    mod given {

        use git::RepoDetails;
        use git_next_core::server::RepoListenUrl;

        use super::*;

        pub fn commit_states(
            states: &[GithubState],
            net: &MockNet,
            repo_details: &git::RepoDetails,
            commit: &git::Commit,
        ) {
            net.on()
                .get(a_commit_status_url(repo_details, commit))
                .respond(StatusCode::OK)
                .body(
                    json!(states
                        .iter()
                        .map(|state| GithubStatus {
                            state: state.to_owned()
                        })
                        .collect::<Vec<_>>())
                    .to_string(),
                )
                .expect("mock");
        }

        pub fn a_commit_status_url(
            repo_details: &git::RepoDetails,
            commit: &git::Commit,
        ) -> String {
            let hostname = repo_details.forge.hostname();
            let repo_path = &repo_details.repo_path;
            format!("https://api.{hostname}/repos/{repo_path}/commits/{commit}/statuses")
        }

        pub fn a_webhook_auth() -> WebhookAuth {
            WebhookAuth::generate()
        }

        pub enum Header {
            Valid(WebhookAuth, Body),
            Missing,
            Invalid,
        }
        pub fn a_webhook_message(header: Header) -> ForgeNotification {
            let body = match &header {
                Header::Valid(_, body) => body.clone(),
                _ => given::a_webhook_message_body(),
            };
            ForgeNotification::new(
                given::a_forge_alias(),
                given::a_repo_alias(),
                given::webhook_headers(header),
                body,
            )
        }
        pub fn webhook_headers(header: Header) -> BTreeMap<String, String> {
            let mut headers = BTreeMap::new();
            match header {
                Header::Valid(auth, body) => {
                    if let Some(sig) = crate::webhook::sign_body(&auth, &body) {
                        headers.insert("x-hub-signature-256".to_string(), format!("sha256={sig}"));
                    }
                }
                Header::Missing => { /* don't add any header */ }
                Header::Invalid => {
                    headers.insert(
                        "x-hub-signature-256".to_string(),
                        format!("{}", WebhookAuth::generate()),
                    );
                }
            }
            headers
        }
        pub fn a_webhook_message_body() -> Body {
            Body::new(a_name())
        }
        pub fn a_commit() -> git::Commit {
            git::Commit::new(
                git::commit::Sha::new(a_name()),
                git::commit::Message::new(a_name()),
            )
        }

        pub fn repo_branches() -> RepoBranches {
            RepoBranches::new(a_name(), a_name(), a_name())
        }

        pub fn a_github_forge(
            repo_details: &git::RepoDetails,
            net: impl Into<kxio::net::Net>,
        ) -> Github {
            Github::new(repo_details.clone(), net.into())
        }

        pub fn repo_details() -> git::RepoDetails {
            git::RepoDetails::new(
                git::Generation::default(),
                &a_repo_alias(),
                &ServerRepoConfig::new(
                    format!("{}/{}", a_name(), a_name()), // repo path: owner/repo
                    a_name(),
                    None,
                    Some(a_name()),
                    Some(a_name()),
                    Some(a_name()),
                ),
                &a_forge_alias(),
                &ForgeConfig::new(
                    ForgeType::ForgeJo,
                    a_name(),
                    a_name(),
                    a_name(),
                    maybe_a_number(),
                    BTreeMap::default(),
                ),
                GitDir::new(PathBuf::default(), StoragePathType::External),
            )
        }

        pub fn a_forge_alias() -> ForgeAlias {
            ForgeAlias::new(a_name())
        }

        pub fn a_repo_alias() -> RepoAlias {
            RepoAlias::new(a_name())
        }
        pub fn a_network() -> kxio::net::MockNet {
            kxio::net::mock()
        }

        pub fn a_repo_listen_url(repo_details: &RepoDetails) -> RepoListenUrl {
            RepoListenUrl::new((
                ListenUrl::new(a_name()),
                repo_details.forge.forge_alias().clone(),
                repo_details.repo_alias.clone(),
            ))
        }
        pub fn any_webhook_url() -> RepoListenUrl {
            given::a_repo_listen_url(&given::repo_details())
        }

        pub fn a_name() -> String {
            use rand::Rng;
            use std::iter;

            fn generate(len: usize) -> String {
                const CHARSET: &[u8] = b"ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789";
                let mut rng = rand::thread_rng();
                let one_char = || CHARSET[rng.gen_range(0..CHARSET.len())] as char;
                iter::repeat_with(one_char).take(len).collect()
            }
            generate(5)
        }

        pub fn maybe_a_number() -> Option<u32> {
            use rand::Rng;
            let mut rng = rand::thread_rng();
            if Rng::gen_ratio(&mut rng, 1, 2) {
                Some(a_number())
            } else {
                None
            }
        }

        pub fn a_number() -> u32 {
            use rand::Rng;
            let mut rng = rand::thread_rng();
            rng.gen_range(0..100)
        }

        pub fn a_webhook_id() -> WebhookId {
            WebhookId::new(given::a_name())
        }

        pub fn a_github_webhook_id() -> i64 {
            rand::thread_rng().next_u32().into()
        }
    }
}
